// ======================================================================
// Nikon camera remote control.
//
// Compatible with the ML-L3 remote control. Supported cameras include:
// D40, D40X, D50, D60, D70, D70s, D80, Coolpix 8400 8800.
//
// Based on an idea from: http://www.bigmike.it/ircontrol/
//
// This software was written for an ATtiny13V with the default fuse
// settings:
//    hfuse = 0xff
//    lfuse = 0x6a
//
// For best results, the ATtiny13 should be calibrated so that the IR
// carrier frequency is as close to 38400 Hz as possible over the expected
// operating voltage of 2.5V..3V. This involves determining the correct OSCCAL
// value for the 9.6MHz oscillator and adapting the macro CAL_BYTE accordingly.
// This value will be close to the factory default, which can be found at
// offset 0 in the calibration space. For my ATtiny13V, I had to increment the
// factory default value by 1. When the CALIBRATE input is low, a continuous
// 38400 Hz signal will be output on the TEST output for 10 seconds after each
// button press.
//
// Power consumption at 3V:
//    powerdown:              0.2 uA    (button not pressed)
//    powerdown + 32 ms WDT:  4.9 uA    (button pressed)
//    idle:                   840 uA
//    LED current:             46 mA
//    RESET supply current     70 uA	(from datasheet)
//
// Approximate power consumption for one shot:
//    Idle current:       156 ms @ 840 uA = 0.036 mAh
//    IR LED:        50% of 7 ms @  46 mA = 0.045 mAh
//                                  total = 0.081 mAh
//
// The capacity of a CR2032 battery of approximately 200 mAh is enough for:
// * 2000 shots
// * 100 years standby (ignoring battery self-discharge)
// * holding the button for 4 years
//
// Copyright 2008 Dick Streefland
//
// This is free software, licensed under the terms of the GNU General
// Public License as published by the Free Software Foundation.
// ======================================================================

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>

#define	OSCILLATOR	9600000
#define	PRESCALER	2
#define	IR_FREQ		38400
#define	CAL_BYTE	0x4e
#define	DEBOUNCE	20		// msec

#define	CLOCK		((OSCILLATOR) >> (PRESCALER))
#define	CYCLES_PER_INT	(((CLOCK)+(IR_FREQ))/(2*(IR_FREQ)))
#define	USEC_PER_PERIOD	((1000000+(IR_FREQ)/2)/(IR_FREQ))

#define	PORT		PORTB
#define	PIN		PINB
#define	DDR		DDRB
#define	LED		0		// PB0 - pin 5
#define	TEST		1		// PB1 - pin 6
#define	NC		2		// PB2 - pin 7
#define	FIRE		3		// PB3 - pin 2
#define	CALIBRATE	4		// PB4 - pin 3

typedef	unsigned char	byte_t;
typedef	unsigned int	uint_t;

// ----------------------------------------------------------------------
// Pattern definition as on/off pairs in usec
// ----------------------------------------------------------------------
static	uint_t	pattern [] PROGMEM =
{
	2000u,	27800u,
	 500u,	 1500u,
	 500u,	 3500u,
	 500u,	63000u,

	2000u,	27800u,
	 500u,	 1500u,
	 500u,	 3500u,
	 500u,	    0u,

	0
};

// ----------------------------------------------------------------------
// Sleep functions
// ----------------------------------------------------------------------
static	void	sleep_idle ( void )
{
	__asm__ __volatile__( "sleep" );	// sleep until next timer interrupt
}

static	void	sleep_powerdown ( void )
{
	MCUCR |= _BV(SM1);			// select power-down sleep mode
	__asm__ __volatile__( "sleep" );	// power down until next interrupt
	MCUCR &= ~ _BV(SM1);			// select idle sleep mode
}

// ----------------------------------------------------------------------
// Switch the IR LED on or off for the specified number of microseconds
// ----------------------------------------------------------------------
static	void	led ( uint_t usec, char on )
{
	uint_t	iter;

	iter = usec / USEC_PER_PERIOD;
	while	( iter-- )
	{
		sleep_idle();
		if	( on )
		{
			PORT |= _BV(LED);
		}
		sleep_idle();
		if	( on )
		{
			PORT &= ~ _BV(LED);
		}
	}
}

// ----------------------------------------------------------------------
// Poll and debounce button
// ----------------------------------------------------------------------
static	_Bool	button ( void )
{
	byte_t	i;

	for	( i = 0; i < 10 * DEBOUNCE; i++ )
	{
		if	( PIN & _BV(FIRE) )
		{
			return 0;	// button not pressed
		}
		led( 100u, 0 );
	}
	return 1;			// button pressed for DEBOUNCE msec
}

// ----------------------------------------------------------------------
// Calibration loop: generate IR_FREQ Hz signal for a number of seconds
// ----------------------------------------------------------------------
static	void	calibrate ( byte_t sec )
{
	uint_t	iter;

	while	( sec-- )
	{
		iter = 1000000 / USEC_PER_PERIOD;
		while	( iter-- )
		{
			sleep_idle();
			PORT |= _BV(TEST);
			sleep_idle();
			PORT &= ~ _BV(TEST);
		}
	}
}

// ----------------------------------------------------------------------
// Interrupt handlers
// ----------------------------------------------------------------------
__attribute__((signal,naked))
extern	void	SIG_OUTPUT_COMPARE0A ( void )
{
	__asm__ __volatile__( "reti" );
}

__attribute__((signal,naked))
extern	void	SIG_PIN_CHANGE0 ( void )
{
	__asm__ __volatile__( "reti" );
}

__attribute__((signal,naked))
extern	void	SIG_WATCHDOG_TIMEOUT ( void )
{
	__asm__ __volatile__( "reti" );
}

// ----------------------------------------------------------------------
// Main
// ----------------------------------------------------------------------
extern	int	main ( void )
{
	uint_t*	p;
	uint_t	usec;

	// Disable watchdog
	MCUSR &= ~ _BV(WDRF);
	WDTCR |= _BV(WDCE);
	WDTCR = 0;

	// Load calibration byte
	OSCCAL = CAL_BYTE;

	// Set clock division factor
	CLKPR = _BV(CLKPCE);
	CLKPR = PRESCALER;

	// Configure I/O pins
	DDR   |= _BV(LED);		// LED output
	DDR   |= _BV(TEST);		// calibration test output
	PORT  |= _BV(NC);		// unused input with pull-up
	PORT  |= _BV(FIRE);		// button input with pull-up
	PORT  |= _BV(CALIBRATE);	// calibration input with pull-up

	// Configure pin change interrupt
	PCMSK = _BV(FIRE);		// select pin for change interrupt
	GIMSK |= _BV(PCIE);		// enable pin change interrupt

	// Configure timer
	TCCR0A = 0x02;			// CTC mode
	TCCR0B = 0x01;			// no scaling
	OCR0A  = CYCLES_PER_INT;	// set timer period
	TIMSK0 |= _BV(OCIE0A);		// enable output compare interrupt
	MCUCR  |= _BV(SE);		// sleep enable

	// Enable interrupts
	sei();

	// Check calibration input
	if	( ! (PIN & _BV(CALIBRATE)) )
	{
		calibrate( 10 );	// generate calibration signal for 10 seconds
	}

	// Wait for button release, with disabled pull-up to save power
	cli();
	WDTCR |= _BV(WDCE);		// prepare for prescale select
	WDTCR = _BV(WDP0);		// 32 ms time-out
	sei();
	WDTCR |= _BV(WDTIE);		// enable watchdog interrupt
	while	( ! (PIN & _BV(FIRE)) )
	{
		PORT &= ~ _BV(FIRE);	// disable pull-up
		sleep_powerdown();
		PORT |= _BV(FIRE);	// re-enable pull-up
	}
	WDTCR &= ~ _BV(WDTIE);		// disable watchdog interrupt

	// Wait for fire button
	while	( ! button() )
	{
		sleep_powerdown();
	}

	// Send IR pattern
	for	( p = pattern; usec = pgm_read_word(p), usec != 0; p += 2 )
	{
		led( usec, 1 );
		usec = pgm_read_word(p + 1);
		led( usec, 0 );
	}

	// Force RESET
	cli();
	WDTCR |= _BV(WDCE);
	WDTCR = _BV(WDE);		// enable watchdog RESET (16 ms)
	for	( ;; )
	{
		sleep_powerdown();
	}

	return 0;
}
